# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A translation module of Ariel package from the Google EMEA gTech Ads Data Science."""

import datetime
import os
import re
from typing import Final, Mapping, Sequence
import tensorflow as tf
from vertexai.generative_models import GenerativeModel

_TRANSLATION_PROMPT: Final[str] = (
    "You're hired by a company called: {}. The received transcript is: {}."
    " Specific instructions: {}. The target language is: {}."
)
_BREAK_MARKER: Final[str] = "<BREAK>"
_DONT_TRANSLATE_MARKER: Final[str] = "<DO NOT TRANSLATE>"


def generate_script(*, utterance_metadata, key: str = "text") -> str:
  """Generates a script string from a list of utterance metadata.

  Args:
    utterance_metadata: The sequence of mappings, where each mapping represents
      utterance metadata with "text", "start", "end", "speaker_id",
      "ssml_gender", "for_dubbing" keys.
    key: The key to use when searching for the strings for the script.

  Returns:
    A string representing the script, with "<BREAK>" inserted
    between chunks.
  """
  trimmed_lines = [
      item[key].strip() if item[key] else "" for item in utterance_metadata
  ]
  return _BREAK_MARKER + _BREAK_MARKER.join(trimmed_lines) + _BREAK_MARKER


def translate_script(
    *,
    script: str,
    advertiser_name: str,
    translation_instructions: str,
    target_language: str,
    model: GenerativeModel,
) -> str:
  """Translates the provided transcript to the target language using a Generative AI model.

  Args:
      script: The transcript to translate.
      advertiser_name: The name of the advertiser.
      translation_instructions: Specific instructions for the translation.
      target_language: The target language for the translation.
      model: The GenerativeModel to use for translation.

  Returns:
      The translated script.
  """
  prompt = _TRANSLATION_PROMPT.format(
      advertiser_name, script, translation_instructions, target_language
  )
  response = model.generate_content(prompt)
  return response.text


class GeminiTranslationError(Exception):
  """Error when Gemini can't translate script correctly."""

  pass


def add_translations(
    *,
    utterance_metadata: Sequence[Mapping[str, str | float]],
    translated_script: str,
) -> Sequence[Mapping[str, str | float]]:
  """Updates the "translated_text" field of each utterance metadata with the corresponding text segment.

  Args:
      utterance_metadata: The sequence of mappings, where each mapping
        represents utterance metadata with "text", "start", "end", "speaker_id",
        "ssml_gender", "path", "for_dubbing" and optionally "vocals_path" keys.
      translated_script: The string containing the translated text segments,
        separated by _BREAK_MARKER.

  Returns:
      A list of updated utterance metadata with the "translated_text" field
      populated.

  Raises:
      ValueError: If the number of utterance metadata and text segments do not
      match.
  """
  stripped_translation = re.sub(
      rf"^\s*{_BREAK_MARKER}\s*|\s*{_BREAK_MARKER}\s*$", "", translated_script
  )
  text_segments = stripped_translation.split(_BREAK_MARKER)
  if len(utterance_metadata) != len(text_segments):
    raise GeminiTranslationError(
        "The utterance metadata must be of the same length as the text"
        f" segments. Currently they are: {len(utterance_metadata)} and"
        f" {len(text_segments)}."
    )
  updated_utterance_metadata = []
  for metadata, translated_text in zip(utterance_metadata, text_segments):
    if translated_text != _DONT_TRANSLATE_MARKER:
      updated_utterance_metadata.append(
          {**metadata, "translated_text": translated_text}
      )
    else:
      continue
  return updated_utterance_metadata


def save_srt_subtitles(
    *,
    utterance_metadata: Sequence[Mapping[str, str | float]],
    output_directory: str,
    target_language: str,
) -> str:
  """Returns a path to an SRT subtitle file from utterance metadata.

  Args:
    utterance_metadata: A list of dictionaries, where each dictionary contains
      information about an utterance, including 'start', 'end', and
      'translated_text'.
    output_directory: The directory where the SRT file will be saved.
    target_language: The language to dub the ad into. It must be ISO 3166-1
      alpha-2 country code.
  """
  srt_content = ""
  for i, utterance in enumerate(utterance_metadata):
    start_time = str(datetime.timedelta(seconds=utterance["start"]))[:-3]
    end_time = str(datetime.timedelta(seconds=utterance["end"]))[:-3]
    start_time = start_time.replace(".", ",").zfill(12)
    end_time = end_time.replace(".", ",").zfill(12)
    srt_content += f"{i+1}\n"
    srt_content += (
        f"{start_time.replace('.', ',')} --> {end_time.replace('.', ',')}\n"
    )
    srt_content += f"{utterance['translated_text']}\n\n"
    target_language_suffix = "_" + target_language.replace("-", "_").lower()
  srt_file_path = os.path.join(
      output_directory, f"translated_subtitles{target_language_suffix}.srt"
  )
  with tf.io.gfile.GFile(srt_file_path, "w") as subtitles_file:
    subtitles_file.write(srt_content)
  return srt_file_path



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module streamlining Ariel runs in Google Colab only."""

import dataclasses
import datetime
import os
import re
import time
from typing import Any, Final, Mapping, Sequence
from absl import logging
from google.auth import default
from googleapiclient.discovery import build
from googleapiclient.discovery import Resource
import gspread
from gspread_dataframe import set_with_dataframe
from IPython.display import clear_output
import pandas as pd
import tensorflow as tf

_BASE_DIRECTORY_COLAB: Final[str] = "/content"
_BASE_DIRECTORY_DRIVE: Final[str] = "/content/drive"
_STRING_COLUMNS: Final[Sequence[str]] = (
    "text",
    "translated_text",
    "speaker_id",
    "ssml_gender",
    "assigned_voice",
)
_FLOAT_COLUMNS: Final[Sequence[str]] = (
    "start",
    "end",
    "pitch",
    "speed",
    "volume_gain_db",
    "stability",
    "similarity_boost",
    "style",
)
_BOOL_COLUMNS: Final[Sequence[str]] = (
    "for_dubbing",
    "adjust_speed",
    "use_speaker_boost",
)


def extract_file_id(sharable_link: str) -> str | None:
  """Extracts the file ID from a sharable link.

  Args:
    sharable_link: The sharable link to the file.

  Returns:
    The file ID, or None if the ID could not be extracted.
  """
  try:
    if "id=" in sharable_link:
      return sharable_link.split("id=")[1]
    elif "/d/" in sharable_link:
      return sharable_link.split("/d/")[1].split("/")[0]
    elif "/file/d/" in sharable_link:
      return sharable_link.split("/file/d/")[1].split("/")[0]
    else:
      return
  except Exception:
    return


def get_parent_path(service: Resource, parent_id: str) -> str:
  """Recursively retrieves the path of parent folders.

  Args:
    service: The Drive API service instance.
    parent_id: The ID of the parent folder.

  Returns:
    The path of the parent folders.
  """
  parent_metadata = (
      service.files()
      .get(fileId=parent_id, fields="name,parents", supportsAllDrives=True)
      .execute()
  )
  parent_path = f"/{parent_metadata['name']}"
  if "parents" in parent_metadata:
    parent_path = (
        get_parent_path(service, parent_metadata["parents"][0]) + parent_path
    )
  return parent_path


def get_file_path_from_sharable_link(sharable_link: str) -> str | None:
  """Determines the file path on Google Drive given a sharable link.

  Args:
    sharable_link: The sharable link to the file.

  Returns:
    The file path on Google Drive, or None if the path could not be determined.
  """
  try:
    file_id = extract_file_id(sharable_link)
    if not file_id:
      return
    creds, _ = default()
    service = build("drive", "v3", credentials=creds)
    metadata = (
        service.files()
        .get(fileId=file_id, fields="name,parents", supportsAllDrives=True)
        .execute()
    )
    path = _BASE_DIRECTORY_DRIVE
    if "parents" in metadata:
      parent_id = metadata["parents"][0]
      path += get_parent_path(service, parent_id)
    path = f'{path}/{metadata["name"]}'
    return path
  except Exception as e:
    logging.error(f"Could not determine the file path due to the error: {e}")
    return


def copy_file_to_colab(
    *, source_file_path: str, destination_folder: str = _BASE_DIRECTORY_COLAB
) -> str:
  """Returns a file path to the copied input file.

  Args:
    source_file_path: The path to the source file.
    destination_folder: The destination directory in Colab.
  """
  destination_file_path = (
      f"{destination_folder}/{os.path.basename(source_file_path)}"
  )
  tf.io.gfile.copy(source_file_path, destination_file_path, overwrite=True)
  return destination_file_path


def copy_output_to_google_drive(
    *, colab_dir: str, google_drive_dir: str
) -> str:
  """Copies the contents of the 'output' subdirectory in a Colab directory.

  to a new directory in Google Drive with the same name as the Colab directory.

  Args:
    colab_dir: The path to the Colab directory.
    google_drive_dir: The path to the Google Drive directory.

  Returns:
    The path to the Google Drive destination directory.
  """
  output_dir = os.path.join(colab_dir, "output")
  destination_dir = os.path.join(google_drive_dir, os.path.basename(colab_dir))
  tf.io.gfile.makedirs(destination_dir)
  for filename in os.listdir(output_dir):
    source_path = os.path.join(output_dir, filename)
    destination_path = os.path.join(destination_dir, filename)
    if not tf.io.gfile.exists(destination_path):
      tf.io.gfile.copy(source_path, destination_path, overwrite=True)
  return destination_dir


def get_google_sheet_as_dataframe(sheet_link: str) -> pd.DataFrame:
  """Returns Google Sheet using the provided URL as a Pandas DataFrame.

  Args:
    sheet_link: The URL of the Google Sheet.

  Returns:
    A Pandas DataFrame containing the sheet data.
  """
  creds, _ = default()
  gc = gspread.authorize(creds)
  spreadsheet = gc.open_by_url(sheet_link)
  worksheet = spreadsheet.sheet1
  rows = worksheet.get_all_values()
  df = pd.DataFrame(rows[1:], columns=rows[0])
  df.columns = df.columns.str.replace(r"\s+", "", regex=True)
  return df


@dataclasses.dataclass
class ScriptMetadata:
  """Instance with script metadata.

  Attributes:
      script_with_timestamps: A sequence of dictionaries, where each dictionary
        represents a speech segment and contains the keys "start", "end",
        "text", "speaker_id", "ssml_gender", "assigned_voice" and "adjust_speed"
      google_text_to_speech_parameters: A sequence of mappings with Google
        Text-to-Speech parameters ("pitch", "speed", "volume_gain_db"), or None
        if not applicable.
      elevenlabs_text_to_speech_parameters: A sequence of mappings with
        ElevenLabs Text-to-Speech parameters ("stability", "similarity_boost",
        "style", "use_speaker_boost"), or None if not applicable.
  """

  script_with_timestamps: Sequence[Mapping[str, float | str]]
  google_text_to_speech_parameters: (
      Sequence[Mapping[str, float | int]] | None
  ) = None
  elevenlabs_text_to_speech_parameters: (
      Sequence[Mapping[str, float | int | bool]] | None
  ) = None


def create_script_metadata_from_dataframe(df: pd.DataFrame) -> ScriptMetadata:
  """Returns a ScriptMetadata instance from a Pandas DataFrame.

  Args:
    df: The DataFrame to process.
  """
  script_with_timestamps = [
      {
          "start": start,
          "end": end,
          "text": text,
          "speaker_id": speaker_id,
          "ssml_gender": ssml_gender,
          "assigned_voice": assigned_voice,
          "adjust_speed": adjust_speed,
      }
      for start, end, text, speaker_id, ssml_gender, assigned_voice, adjust_speed in zip(
          df["start"],
          df["end"],
          df["text"],
          df["speaker_id"],
          df["ssml_gender"],
          df["assigned_voice"],
          df["adjust_speed"],
      )
  ]
  metadata_kwargs = {
      "script_with_timestamps": script_with_timestamps,
  }
  if {"stability", "similarity_boost", "style", "use_speaker_boost"} <= set(
      df.columns
  ):
    metadata_kwargs["elevenlabs_text_to_speech_parameters"] = [
        {
            "stability": stability,
            "similarity_boost": similarity_boost,
            "style": style,
            "use_speaker_boost": use_speaker_boost,
        }
        for stability, similarity_boost, style, use_speaker_boost in zip(
            df["stability"],
            df["similarity_boost"],
            df["style"],
            df["use_speaker_boost"],
        )
    ]
  elif {"pitch", "speed", "volume_gain_db"} <= set(df.columns):
    metadata_kwargs["google_text_to_speech_parameters"] = [
        {
            "pitch": pitch,
            "speed": speed,
            "volume_gain_db": volume_gain_db,
        }
        for pitch, speed, volume_gain_db in zip(
            df["pitch"], df["speed"], df["volume_gain_db"]
        )
    ]
  return ScriptMetadata(**metadata_kwargs)


def convert_utterance_metadata(
    utterance_metadata: pd.DataFrame,
) -> pd.DataFrame:
  """Converts specific columns in utterance metadata.

  Expected columns:
      Always present:
          - text: The original text of the utterance.
          - translated_text: The translated text of the utterance.
          - speaker_id: An identifier for the speaker.
          - ssml_gender: The gender specified in SSML for the speaker.
          - assigned_voice: The voice assigned to the speaker.
          - start: The start time of the utterance in seconds.
          - end: The end time of the utterance in seconds.
          - for_dubbing: Whether the utterance is for dubbing.
          - adjust_speed: Whether to adjust the speed of the utterance.

      Optional:
          Case 1:
              - pitch: The pitch adjustment for the utterance.
              - speed: The speed adjustment for the utterance.
              - volume_gain: The volume gain for the utterance.

          Case 2:
              - stability: A stability value for the utterance (e.g., for voice
              synthesis).
              - similarity_boost: A similarity boost value (e.g., for voice
              synthesis).
              - style: A style value (e.g., for voice synthesis).
              - use_speaker_boost: Whether to use speaker boost.

  Args:
    utterance_metadata: The input utterance metadata with the expected columns.

  Returns:
    The converted utterance metadata with the correct data types.
  """
  for col in _STRING_COLUMNS:
    if col in utterance_metadata.columns:
      utterance_metadata[col] = utterance_metadata[col].astype(str)
  for col in _FLOAT_COLUMNS:
    if col in utterance_metadata.columns:
      utterance_metadata[col] = utterance_metadata[col].astype(float)
  for col in _BOOL_COLUMNS:
    if col in utterance_metadata.columns:
      utterance_metadata[col] = utterance_metadata[col].str.lower() == "true"
  return utterance_metadata


def get_folder_id_by_path(path: str) -> str:
  """Returns the Google Drive folder ID for a specified path.

  Args:
    path: The full path of the folder in Google Drive, starting from
      '/content/drive/My Drive/...'. For example: '/content/drive/My
      Drive/parent_folder/sub_folder'.

  Raises:
    FileNotFoundError: If any part of the specified path does not exist in
    Google Drive.
  """
  path_parts = path.split("/")[4:]
  folder_id = "root"
  creds, _ = default()
  service = build("drive", "v3", credentials=creds)
  for part in path_parts:
    query = (
        f"'{folder_id}' in parents and name = '{part}' and mimeType ="
        " 'application/vnd.google-apps.folder' and trashed = false"
    )
    results = service.files().list(q=query, fields="files(id, name)").execute()
    items = results.get("files", [])
    if not items:
      raise FileNotFoundError(f"Folder '{part}' not found in the current path.")
    folder_id = items[0]["id"]
  return folder_id


def save_dataframe_to_gdrive(
    *, dataframe: str, google_drive_directory_id: str, sheet_name: str
) -> None:
  """Saves a DataFrame as a Google Sheet in a specified folder.

  Args:
    dataframe: The DataFrame to be saved as a Google Sheet.
    google_drive_directory_id: The ID of the Google Drive folder where the
      Google Sheet should be saved.
    sheet_name: The desired name of the Google Sheet.
  """
  creds, _ = default()
  service = build("drive", "v3", credentials=creds)
  client = gspread.authorize(creds)
  file_metadata = {
      "name": sheet_name,
      "mimeType": "application/vnd.google-apps.spreadsheet",
      "parents": [google_drive_directory_id],
  }
  file = service.files().create(body=file_metadata, fields="id").execute()
  spreadsheet_id = file.get("id")
  spreadsheet = client.open_by_key(spreadsheet_id)
  worksheet = spreadsheet.get_worksheet(0)
  set_with_dataframe(worksheet, dataframe)


def convert_utterance_metadata_to_google_sheets(
    *,
    input_file_google_drive_path: str,
    output_directory: str,
    wait: bool = True,
    wait_time: int = 10,
    remove_json: bool = True,
) -> None:
  """Converts utterance metadata from JSON files to Google Sheets.

  Args:
    input_file_google_drive_path: The Google Drive path of the input directory
      containing JSON files.
    output_directory: The output directory in Google Drive where the Google
      Sheets will be saved.
    wait: Whether to wait for JSON files to appear in Google Drive.
    wait_time: How long to wait for all the files to appear on Google Drive.
    remove_json: Whether to remove the original JSON files after conversion.
      Defaults to True.
  """
  google_drive_directory = os.path.join(
      os.path.split(input_file_google_drive_path)[0],
      os.path.split(output_directory)[1],
  )
  if wait:
    logging.info(
        f"Waiting {wait_time}s for all the files to appear on Google Drive."
    )
    time.sleep(wait_time)
  google_drive_directory_id = get_folder_id_by_path(google_drive_directory)
  json_files = [
      file
      for file in tf.io.gfile.listdir(google_drive_directory)
      if file.endswith(".json")
  ]
  if not json_files:
    logging.info(
        f"No JSON files found in the directory: {google_drive_directory}"
    )
    return
  json_paths = [
      os.path.join(google_drive_directory, file) for file in json_files
  ]
  spreadsheet_names = [os.path.splitext(file)[0] for file in json_files]
  for json_path, spreadsheet_name in zip(json_paths, spreadsheet_names):
    utterance_metadata_df = pd.read_json(json_path)
    save_dataframe_to_gdrive(
        dataframe=utterance_metadata_df,
        google_drive_directory_id=google_drive_directory_id,
        sheet_name=spreadsheet_name,
    )
    if remove_json:
      tf.io.gfile.remove(json_path)


@dataclasses.dataclass
class ColabPaths:
  """Instance with Colab file paths.

  Attributes:
    input_file_google_drive_path: The Google Drive path of the input file.
    input_file_colab_path: The path to the input file in Colab after copying
      from Google Drive.
    audio_file_colab_path: An optional path to a file with the audio part only.
      It should be vocals + background audio or just background audio. It will
      be used instead of the audio track from the input video. Must be an MP3
      file.
    vocals_file_colab_path: The path to the vocals file in Colab after copying
      from Google Drive, or None if not provided.
    background_file_colab_path: The path to the background file in Colab after
      copying from Google Drive, or None if not provided.
  """

  input_file_google_drive_path: str
  input_file_colab_path: str | None
  audio_file_colab_path: str | None = None
  vocals_file_colab_path: str | None = None
  background_file_colab_path: str | None = None


def generate_colab_file_paths(
    *,
    video_google_drive_link: str,
    audio_google_drive_link: str | None = None,
    vocals_google_drive_link: str | None = None,
    background_google_drive_link: str | None = None,
) -> ColabPaths:
  """Generates Colab file paths for the specified Google Drive links and copies files to Colab.

  Args:
      video_google_drive_link: The Google Drive link to the main input file.
      audio_google_drive_link: The Google Drive link to the audio file, if
        available. Defaults to None.
      vocals_google_drive_link: The Google Drive link to the vocals file, if
        available. Defaults to None.
      background_google_drive_link: The Google Drive link to the background
        file, if available. Defaults to None.

  Returns:
      ColabPaths: An instance of ColabPaths containing the Google Drive and
      Colab paths for each file.
  """
  input_file_google_drive_path = get_file_path_from_sharable_link(
      video_google_drive_link
  )
  input_file_colab_path = copy_file_to_colab(
      source_file_path=input_file_google_drive_path
  )
  audio_file_colab_path = None
  if audio_google_drive_link:
    audio_file_google_drive_path = get_file_path_from_sharable_link(
        audio_google_drive_link
    )
    audio_file_colab_path = copy_file_to_colab(
        source_file_path=audio_file_google_drive_path
    )
  vocals_file_colab_path = None
  if vocals_google_drive_link:
    vocals_file_google_drive_path = get_file_path_from_sharable_link(
        vocals_google_drive_link
    )
    vocals_file_colab_path = copy_file_to_colab(
        source_file_path=vocals_file_google_drive_path
    )
  background_file_colab_path = None
  if background_google_drive_link:
    background_file_google_drive_path = get_file_path_from_sharable_link(
        background_google_drive_link
    )
    background_file_colab_path = copy_file_to_colab(
        source_file_path=background_file_google_drive_path
    )
  return ColabPaths(
      input_file_google_drive_path=input_file_google_drive_path,
      input_file_colab_path=input_file_colab_path,
      audio_file_colab_path=audio_file_colab_path,
      vocals_file_colab_path=vocals_file_colab_path,
      background_file_colab_path=background_file_colab_path,
  )


def _generate_default_output_folder(advertiser_name: str) -> str:
  """Generates a default output folder path.

  The function removes any non-alphabet characters from the advertiser's name,
  converts it to lowercase,
  and appends a timestamp (in the format YYYYMMDDHHMMSSffffff) to ensure the
  folder name is unique.

  Args:
      advertiser_name: The name of the advertiser used to create the folder
        name.

  Returns:
    A string representing the full path of the output folder, starting with
    '/content' and ending
    with the formatted advertiser name and timestamp.
  """
  formatted_advertiser_name = re.sub(r"[^a-zA-Z]", "", advertiser_name).lower()
  now = datetime.datetime.now()
  return os.path.join(
      "/content",
      "dubbing_" + formatted_advertiser_name + now.strftime("%Y%m%d%H%M%S%f"),
  )


def setup_output_folder(
    *,
    advertiser_name: str,
    input_file_google_drive_path: str,
    output_folder: str | None = None,
    metadata_google_drive_link: str | None = None,
) -> str:
  """Returns output folder path, either based on user input or generated automatically.

  If `output_folder` is not provided, a default folder name will be generated
  using the advertiser's
  name and a timestamp. If a folder with the specified name already exists in
  Google Drive, the user
  will be prompted to overwrite it or enter a new name. If the user leaves the
  input blank, a default
  folder name will be created automatically.

  Args:
      advertiser_name: The name of the advertiser, used for generating a default
        folder name.
      input_file_google_drive_path: The path of the input file in Google Drive.
      output_folder: An optional specified output folder name. Defaults to None.
      metadata_google_drive_link: A link to a utterance metadata Google Sheet.

  Returns:
      The path of the created output folder.
  """
  if not output_folder:
    output_folder = _generate_default_output_folder(advertiser_name)
  google_drive_output_path = os.path.join(
      os.path.split(input_file_google_drive_path)[0], output_folder
  )
  while tf.io.gfile.exists(google_drive_output_path):
    user_response = input(
        f"The folder '{google_drive_output_path}' already exists in your Google"
        " Drive. Do you want to overwrite it? (yes/no): "
    )
    if user_response.lower() == "yes":
      break
    else:
      output_folder = input(
          "Please enter a new output folder name, or leave it empty to"
          " auto-generate one: "
      ).strip()
      if not output_folder:
        output_folder = _generate_default_output_folder(advertiser_name)
      google_drive_output_path = os.path.join(
          os.path.split(input_file_google_drive_path)[0], output_folder
      )
  output_folder = os.path.join("/content", output_folder)
  if metadata_google_drive_link:
    logging.info(
        "You're using utterance metadata from Google Sheets. The Colab output"
        " directory will be cleaned if it exists already."
    )
    try:
      tf.io.gfile.rmtree(output_folder)
    except tf.errors.NotFoundError:
      pass
  tf.io.gfile.makedirs(output_folder)
  return output_folder


def process_dubbing(
    *,
    dubber: Any,
    input_file_google_drive_path: str,
    output_folder: str,
    script_google_drive_link: str | None = None,
    metadata_google_drive_link: str | None = None,
) -> None:
  """Processes the dubbing workflow based on provided Google Drive links for script or metadata.

  This function authenticates the user, retrieves data from Google Sheets, and
  performs the dubbing
  based on the type of metadata provided. If a script link is provided, it
  processes it with voice
  metadata; if only a metadata link is provided, it uses utterance metadata. If
  neither is provided,
  it defaults to a basic dubbing process. Finally, it copies the output to
  Google Drive.

  Args:
      dubber: The dubbing object responsible for dubbing functionality.
      input_file_google_drive_path: The path of the input file in Google Drive.
      output_folder: The output folder in Google Drive where results will be
        saved.
      script_google_drive_link: Google Drive link to the script with voice
        metadata.
      metadata_google_drive_link: Google Drive link to the utterance metadata.
  """
  if script_google_drive_link:
    script_with_voice_metadata_df = convert_utterance_metadata(
        get_google_sheet_as_dataframe(script_google_drive_link)
    )
    script_with_voice_metadata = create_script_metadata_from_dataframe(
        script_with_voice_metadata_df
    )
    _ = dubber.dub_ad_from_script(
        script_with_timestamps=script_with_voice_metadata.script_with_timestamps,
        google_text_to_speech_parameters=script_with_voice_metadata.google_text_to_speech_parameters,
        elevenlabs_text_to_speech_parameters=script_with_voice_metadata.elevenlabs_text_to_speech_parameters,
    )
  elif metadata_google_drive_link:
    utterance_metadata_df = get_google_sheet_as_dataframe(
        metadata_google_drive_link
    )
    converted_utterance_metadata_df = convert_utterance_metadata(
        utterance_metadata_df
    )
    utterance_metadata = converted_utterance_metadata_df.to_dict("records")
    _ = dubber.dub_ad_with_utterance_metadata(
        utterance_metadata=utterance_metadata
    )
  else:
    _ = dubber.dub_ad()
  clear_output(wait=True)
  print(
      "The dubbing process is finished. Copying files to your Google Drive."
      " Please wait..."
  )
  destination_dir = copy_output_to_google_drive(
      colab_dir=output_folder,
      google_drive_dir=os.path.dirname(input_file_google_drive_path),
  )
  convert_utterance_metadata_to_google_sheets(
      input_file_google_drive_path=input_file_google_drive_path,
      output_directory=output_folder,
  )
  print(f"The output is saved in: {destination_dir}")



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An audio processing module of Ariel package from the Google EMEA gTech Ads Data Science."""

import os
import re
import subprocess
from typing import Final
from typing import Final, Mapping, Sequence
from typing import Mapping, Sequence
from absl import logging
import numpy as np
from pyannote.audio import Pipeline
from pydub import AudioSegment
from pyloudnorm import Meter
import tensorflow as tf
import torch

AUDIO_PROCESSING: Final[str] = "audio_processing"
_OUTPUT: Final[str] = "output"
_DEFAULT_DUBBED_VOCALS_AUDIO_FILE: Final[str] = "dubbed_vocals.mp3"
_DEFAULT_DUBBED_AUDIO_FILE: Final[str] = "dubbed_audio"
_DEFAULT_OUTPUT_FORMAT: Final[str] = ".mp3"
_SUPPORTED_DEVICES: Final[tuple[str, str]] = ("cpu", "cuda")
_TIMESTAMP_THRESHOLD: Final[float] = 0.001
_DEFAULT_RATE: Final[float] = 44100
_TARGET_LUFS: Final[float] = -16
_MIN_BLOCK_SIZE_MS: Final[int] = 400


def build_demucs_command(
    *,
    audio_file: str,
    output_directory: str,
    device: str = "cpu",
    shifts: int = 10,
    overlap: float = 0.25,
    mp3_bitrate: int = 320,
    mp3_preset: int = 2,
    jobs: int = 0,
    split: bool = True,
    segment: int | None = None,
    int24: bool = False,
    float32: bool = False,
    flac: bool = False,
    mp3: bool = True,
) -> str:
  """Builds the Demucs audio separation command.

  Args:
      audio_file: The path to the audio file to process.
      output_directory: The output directory for separated tracks.
      device: The device to use ("cuda" or "cpu").
      shifts: The number of random shifts for equivariant stabilization.
      overlap: The overlap between splits.
      mp3_bitrate: The bitrate of converted MP3 files.
      mp3_preset: The encoder preset for MP3 conversion.
      jobs: The number of jobs to run in parallel.
      split: Whether to split audio into chunks.
      segment: The split size for chunks (None for no splitting).
      int24: Save WAV output as 24 bits.
      float32: Save WAV output as float32.
      flac: Convert output to FLAC.
      mp3: Convert output to MP3.

  Returns:
      A string representing the constructed command.

  Raises:
      ValueError: If both int24 and float32 are set to True.
  """

  if int24 and float32:
    raise ValueError("Cannot set both int24 and float32 to True.")
  updated_output_directory = os.path.join(output_directory, AUDIO_PROCESSING)
  command_parts = [
      "python",
      "-m",
      "demucs.separate",
      "-o",
      f"'{updated_output_directory}'",
      "--device",
      device,
      "--shifts",
      str(shifts),
      "--overlap",
      str(overlap),
      "-j",
      str(jobs),
      "--two-stems vocals",
  ]
  if not split:
    command_parts.append("--no-split")
  elif segment is not None:
    command_parts.extend(["--segment", str(segment)])
  if flac:
    command_parts.append("--flac")
  if mp3:
    command_parts.extend([
        "--mp3",
        "--mp3-bitrate",
        str(mp3_bitrate),
        "--mp3-preset",
        str(mp3_preset),
    ])
  if int24:
    command_parts.append("--int24")
  if float32:
    command_parts.append("--float32")
  command_parts.append(f"'{audio_file}'")
  return " ".join(command_parts)


class DemucsCommandError(Exception):
  pass


def execute_demucs_command(command: str) -> None:
  """Executes a Demucs command using subprocess.

  Demucs is a model using AI/ML to detach dialogues
  from the rest of the audio file.

  Args:
      command: The string representing the command to execute.
  """
  try:
    result = subprocess.run(
        command, shell=True, capture_output=True, text=True, check=True
    )
    logging.info(result.stdout)
  except subprocess.CalledProcessError as error:
    logging.warning(
        "Error in the first attempt to separate audio:"
        f" {error}\n{error.stderr}. Retrying with 'python3' instead of"
        " 'python'."
    )
    python3_command = command.replace("python", "python3", 1)
    try:
      result = subprocess.run(
          python3_command,
          shell=True,
          capture_output=True,
          text=True,
          check=True,
      )
      logging.info(result.stdout)
    except subprocess.CalledProcessError as error:
      raise DemucsCommandError(
          f"Error in final attempt to separate audio: {error}\n{error.stderr}"
      )


def extract_command_info(command: str) -> tuple[str, str, str]:
  """Extracts folder name, output file extension, and input file name (without path) from a Demucs command.

  Args:
      command: The Demucs command string.

  Returns:
      tuple: A tuple containing (folder_name, output_file_extension,
      input_file_name).
  """
  folder_pattern = r"-o\s+(['\"]?)(.+?)\1"
  flac_pattern = r"--flac"
  mp3_pattern = r"--mp3"
  int24_or_float32_pattern = r"--(int24|float32)"
  input_file_pattern = r"['\"]?(\w+\.\w+)['\"]?$|\s(\w+\.\w+)$"
  folder_match = re.search(folder_pattern, command)
  flac_match = re.search(flac_pattern, command)
  mp3_match = re.search(mp3_pattern, command)
  int24_or_float32_match = re.search(int24_or_float32_pattern, command)
  input_file_match = re.search(input_file_pattern, command)
  output_directory = folder_match.group(2) if folder_match else ""
  input_file_name_with_ext = (
      (input_file_match.group(1) or input_file_match.group(2))
      if input_file_match
      else ""
  )
  input_file_name_no_ext = (
      os.path.splitext(input_file_name_with_ext)[0] if input_file_match else ""
  )
  if flac_match:
    output_file_extension = ".flac"
  elif mp3_match:
    output_file_extension = ".mp3"
  elif int24_or_float32_match:
    output_file_extension = ".wav"
  else:
    output_file_extension = ".wav"
  return output_directory, output_file_extension, input_file_name_no_ext


def assemble_split_audio_file_paths(command: str) -> tuple[str, str]:
  """Returns paths to the audio files with vocals and no vocals.

    Args:
        command: The Demucs command string.

  Returns:
      A tuple with a path to the file with the audio with vocals only
      and the other with the background sound only.
  """
  output_directory, output_file_extension, input_file_name = (
      extract_command_info(command)
  )
  audio_vocals_file = f"{output_directory}/htdemucs/{input_file_name}/vocals{output_file_extension}"
  audio_background_file = f"{output_directory}/htdemucs/{input_file_name}/no_vocals{output_file_extension}"
  return audio_vocals_file, audio_background_file


def execute_vocals_non_vocals_split(
    *, audio_file: str, output_directory: str, device: str
) -> tuple[str, str]:
  """Splits an audio file into vocal and non-vocal (background) components using Demucs.

  Args:
    audio_file: The path to the input audio file.
    output_directory: The directory where the separated audio files will be
      saved.
    device: The device to use for Demucs processing (e.g., "cuda" for GPU or
      "cpu").

  Returns:
    A tuple containing the paths to the separated audio files:
      - The path to the vocals audio file.
      - The path to the background audio file.
  """
  demucs_command = build_demucs_command(
      audio_file=audio_file,
      output_directory=output_directory,
      device=device,
  )
  audio_vocals_file, audio_background_file = assemble_split_audio_file_paths(
      command=demucs_command
  )
  if tf.io.gfile.exists(audio_vocals_file) and tf.io.gfile.exists(
      audio_background_file
  ):
    logging.info(
        "The DEMUCS command will not be executed, because the expected files"
        f" {audio_vocals_file} and {audio_background_file} already exist."
    )
  else:
    execute_demucs_command(command=demucs_command)
  return audio_vocals_file, audio_background_file


def split_audio_track(
    audio_file: str,
    output_directory: str,
    device: str,
    voice_separation_rounds: int = 2,
) -> tuple[str, str]:
  """Splits an audio track into vocal and non-vocal components, with optional iterative refinement.

  This function separates the vocals from the background music in an audio file.
  It first checks if the separated files already exist to avoid redundant
  processing.
  If not, it uses the `execute_vocal_non_vocals_split` function to perform the
  initial separation.

  To further refine the separation, it can iteratively apply the voice
  separation
  process to the background track. This helps to remove any residual vocal
  traces
  from the background. Finally, it cleans up temporary files generated by
  Demucs.

  Args:
    audio_file: The path to the input audio file.
    output_directory: The directory to store the separated audio files.
    device: The device to use for Demucs processing (e.g., "cuda" or "cpu").
    voice_separation_rounds: The number of times to iteratively apply voice
      separation to the background track (default is 2).

  Returns:
    A tuple containing the paths to the separated audio files:
      - The path to the vocals audio file.
      - The path to the background audio file.
  """
  demucs_command = build_demucs_command(
      audio_file=audio_file,
      output_directory=output_directory,
      device=device,
  )
  audio_vocals_file, _ = assemble_split_audio_file_paths(command=demucs_command)
  file_extension = os.path.splitext(audio_vocals_file)[1]
  vocals_path = os.path.join(
      output_directory, AUDIO_PROCESSING, f"vocals{file_extension}"
  )
  background_path = os.path.join(
      output_directory, AUDIO_PROCESSING, f"no_vocals{file_extension}"
  )
  if tf.io.gfile.exists(vocals_path) and tf.io.gfile.exists(background_path):
    logging.info(
        "The DEMUCS command will not be executed, because the expected files"
        f" {vocals_path} and {background_path} already exist."
    )
    return vocals_path, background_path
  audio_vocals_file, audio_background_file = execute_vocals_non_vocals_split(
      audio_file=audio_file, output_directory=output_directory, device=device
  )
  tf.io.gfile.copy(audio_vocals_file, vocals_path)
  tf.io.gfile.copy(audio_background_file, background_path)
  if voice_separation_rounds > 1:
    for _ in range(1, voice_separation_rounds):
      _, audio_background_file = execute_vocals_non_vocals_split(
          audio_file=background_path,
          output_directory=output_directory,
          device=device,
      )
      tf.io.gfile.copy(audio_background_file, background_path, overwrite=True)
  tf.io.gfile.rmtree(
      os.path.join(output_directory, AUDIO_PROCESSING, "htdemucs")
  )
  return vocals_path, background_path


def prepare_override_audio_files(
    *, vocals_audio_file: str, background_audio_file: str, output_directory: str
) -> Sequence[str]:
  """Prepares override audio files for processing.

  Args:
    vocals_audio_file: Path to the override vocals audio file.
    background_audio_file: Path to the override background audio file.
    output_directory: The path to the output directory.

  Returns:
    A tuple containing the paths to the processed vocals and background audio
    files.

  Raises:
    ValueError: If the input files are not provided or are not in MP3 format.
  """
  if not (vocals_audio_file and background_audio_file):
    raise ValueError(
        "Both `vocals_audio_file` and `background_audio_file` must be provided."
    )
  _, vocals_extension = os.path.splitext(vocals_audio_file)
  _, background_extension = os.path.splitext(background_audio_file)
  if (
      vocals_extension.lower() != ".mp3"
      or background_extension.lower() != ".mp3"
  ):
    raise ValueError(
        "The vocals and background files must be in the MP3 format, now they"
        f" are {vocals_extension} and {background_extension} respectively."
    )
  target_vocals_path = os.path.join(
      output_directory, AUDIO_PROCESSING, "vocals.mp3"
  )
  target_background_path = os.path.join(
      output_directory, AUDIO_PROCESSING, "no_vocals.mp3"
  )
  tf.io.gfile.copy(vocals_audio_file, target_vocals_path, overwrite=True)
  tf.io.gfile.copy(
      background_audio_file, target_background_path, overwrite=True
  )
  return target_vocals_path, target_background_path


def create_pyannote_timestamps(
    *,
    audio_file: str,
    number_of_speakers: int,
    pipeline: Pipeline,
    device: str = "cpu",
) -> Sequence[Mapping[str, float]]:
  """Creates timestamps from a vocals file using Pyannote speaker diarization.

  Args:
      audio_file: The path to the audio file with vocals.
      number_of_speakers: The number of speakers in the vocal audio file.
      pipeline: Pre-loaded Pyannote Pipeline object.
      device: The device to use during the process.

  Returns:
      A list of dictionaries containing start and end timestamps for each
      speaker segment.
  """
  if device not in _SUPPORTED_DEVICES:
    raise ValueError(
        "The device must be either (' or ').join(_SUPPORTED_DEVICES). Got:"
        f" {device}"
    )
  if device == "cuda":
    pipeline.to(torch.device("cuda"))
  diarization = pipeline(audio_file, num_speakers=number_of_speakers)
  utterance_metadata = [
      {"start": segment.start, "end": segment.end}
      for segment, _, _ in diarization.itertracks(yield_label=True)
  ]
  return utterance_metadata


def merge_utterances(
    *,
    utterance_metadata: Sequence[Mapping[str, float]],
    minimum_merge_threshold: float = _TIMESTAMP_THRESHOLD,
) -> Sequence[Mapping[str, str | float]]:
  """Merges utterances that are within the specified timestamp threshold.

  Args:
    utterance_metadata: A sequence of utterance metadata, each represented as a
      dictionary with keys: "start" and "end".
    minimum_merge_threshold: The maximum time difference between the end of one
      utterance and the start of the next for them to be considered mergeable.

  Returns:
    A list of merged utterance metadata.
  """

  merged_utterances = []
  index = 0
  while index < len(utterance_metadata):
    current_utterance = utterance_metadata[index]
    merged_utterance = current_utterance.copy()
    next_index = index + 1
    while (
        next_index < len(utterance_metadata)
        and utterance_metadata[next_index]["start"] - current_utterance["end"]
        < minimum_merge_threshold
    ):
      merged_utterance["end"] = utterance_metadata[next_index]["end"]
      next_index += 1
    merged_utterances.append(merged_utterance)
    index = next_index
  return merged_utterances


def cut_and_save_audio(
    *,
    audio: AudioSegment,
    utterance: Mapping[str, str | float],
    prefix: str,
    output_directory: str,
) -> str:
  """Cuts a specified segment from an audio file, saves it as an MP3, and returns the path of the saved file.

  Args:
      audio: The audio file from which to extract the segment.
      utterance: A dictionary containing the start and end times of the segment
        to be cut. - 'start': The start time of the segment in seconds. - 'end':
        The end time of the segment in seconds.
      prefix: A string to be used as a prefix in the filename of the saved audio
        segment.
      output_directory: The directory path where the cut audio segment will be
        saved.

  Returns:
      The path of the saved MP3 file.
  """
  start_time_ms = int(utterance["start"] * 1000)
  end_time_ms = int(utterance["end"] * 1000)
  chunk = audio[start_time_ms:end_time_ms]
  chunk_filename = f"{prefix}_{utterance['start']}_{utterance['end']}.mp3"
  chunk_path = f"{output_directory}/{AUDIO_PROCESSING}/{chunk_filename}"
  chunk.export(chunk_path, format="mp3")
  return chunk_path


def run_cut_and_save_audio(
    *,
    utterance_metadata: Sequence[Mapping[str, float]],
    audio_file: str,
    output_directory: str,
    elevenlabs_clone_voices: bool = False,
) -> Sequence[Mapping[str, float]]:
  """Cuts an audio file into chunks based on provided time ranges and saves each chunk to a file.

  Args:
      utterance_metadata: The list of dictionaries, each containing 'start' and
        'end' times in seconds.
      audio_file: The path to the audio file to be cut.
      output_directory: The path to the folder where the audio chunks will be
        saved.
      elevenlabs_clone_voices: Whether to clone source voices. It requires using
        ElevenLabs API.

  Returns:
      A list of dictionaries, each containing the path to the saved chunk, and
      the original start and end times.
  """

  audio = AudioSegment.from_file(audio_file)
  key = "vocals_path" if elevenlabs_clone_voices else "path"
  prefix = "vocals_chunk" if elevenlabs_clone_voices else "chunk"
  updated_utterance_metadata = []
  for utterance in utterance_metadata:
    if elevenlabs_clone_voices:
      expected_chunk_filename = (
          f"{prefix}_{utterance['start']}_{utterance['end']}.mp3"
      )
    else:
      expected_chunk_filename = (
          f"{prefix}_{utterance['start']}_{utterance['end']}.mp3"
      )
    chunk_path = (
        f"{output_directory}/{AUDIO_PROCESSING}/{expected_chunk_filename}"
    )
    if not tf.io.gfile.exists(chunk_path):
      chunk_path = cut_and_save_audio(
          audio=audio,
          utterance=utterance,
          prefix=prefix,
          output_directory=output_directory,
      )
    utterance_copy = utterance.copy()
    utterance_copy[key] = chunk_path
    updated_utterance_metadata.append(utterance_copy)
  return updated_utterance_metadata


def verify_added_audio_chunk(
    *,
    audio_file: str,
    utterance: Mapping[str, str | float],
    output_directory: str,
) -> Mapping[str, str | float]:
  """Verifies and processes a newly added audio chunk.

  Args:
      audio_file: The path to the main audio file containing the added chunk.
      utterance: A dictionary describing the start and end times of the added
        chunk.
      output_directory: The directory to save the processed chunk(s).

  Returns:
      A modified copy of the utterance dictionary with added paths to the saved
      audio chunk(s).
        - 'path': The path to the saved audio chunk.
        - 'vocals_path': The path to the saved vocals chunk (if
        `elevenlabs_clone_voices` is True).
  """
  audio = AudioSegment.from_file(audio_file)
  utterance_copy = utterance.copy()
  chunk_path = cut_and_save_audio(
      audio=audio,
      utterance=utterance,
      prefix="chunk",
      output_directory=output_directory,
  )
  utterance_copy["path"] = chunk_path
  return utterance_copy


def verify_modified_audio_chunk(
    *,
    audio_file: str,
    utterance: Mapping[str, str | float],
    output_directory: str,
) -> Mapping[str, str | float]:
  """Verifies and reprocesses a potentially modified audio chunk, potentially including isolated vocals.

  Args:
      audio_file: The path to the main audio file containing the modified chunk.
      utterance: A dictionary describing the start and end times of the modified
        chunk and its expected paths.
      output_directory: The directory to save the processed chunk(s).

  Returns:
      A modified copy of the utterance dictionary with potentially updated paths
      to the saved audio chunk(s).
        - 'path': The updated path to the saved audio chunk (if it was
        modified).
        - 'vocals_path': The updated path to the saved vocals chunk (if
        `elevenlabs_clone_voices` is True and it was modified).
  """
  audio = AudioSegment.from_file(audio_file)
  utterance_copy = utterance.copy()
  expected_chunk_path = f"chunk_{utterance['start']}_{utterance['end']}.mp3"
  actual_chunk_path = utterance_copy["path"]
  if expected_chunk_path != actual_chunk_path:
    tf.io.gfile.remove(actual_chunk_path)
    chunk_path = cut_and_save_audio(
        audio=audio,
        utterance=utterance,
        prefix="chunk",
        output_directory=output_directory,
    )
    utterance_copy["path"] = chunk_path
  return utterance_copy


def insert_audio_at_timestamps(
    *,
    utterance_metadata: Sequence[Mapping[str, str | float]],
    background_audio_file: str,
    output_directory: str,
) -> str:
  """Inserts audio chunks into a background audio track at specified timestamps.

  Args:
    utterance_metadata: A sequence of utterance metadata, each represented as a
      dictionary with keys: "text", "start", "end", "speaker_id", "ssml_gender",
      "translated_text", "assigned_google_voice", "for_dubbing", "path" and
      optionally "vocals_path".
    background_audio_file: Path to the background audio file.
    output_directory: Path to save the output audio file.

  Returns:
    The path to the output audio file.
  """

  background_audio = AudioSegment.from_mp3(background_audio_file)
  total_duration = background_audio.duration_seconds
  output_audio = AudioSegment.silent(duration=total_duration * 1000)
  meter = Meter(rate=_DEFAULT_RATE)
  for item in utterance_metadata:
    start_time = int(item["start"] * 1000)
    if not item["for_dubbing"]:
      audio_chunk = AudioSegment.from_mp3(item["path"])
    else:
      audio_chunk = AudioSegment.from_mp3(item["dubbed_path"])
      if len(audio_chunk) < _MIN_BLOCK_SIZE_MS:
        logging.error(
            f"The dubbed chunk duaration is less than {_MIN_BLOCK_SIZE_MS}."
            f" Silent padding of {_MIN_BLOCK_SIZE_MS} will be added to"
            " normalize the volume."
        )
        padding_duration = _MIN_BLOCK_SIZE_MS - len(audio_chunk)
        audio_chunk += AudioSegment.silent(duration=padding_duration)
      try:
        samples = np.array(audio_chunk.get_array_of_samples())
        samples = samples.astype(np.float32) / np.iinfo(samples.dtype).max
        loudness = meter.integrated_loudness(samples)
        loudness_difference = _TARGET_LUFS - loudness
        audio_chunk = audio_chunk.apply_gain(loudness_difference)
      except ValueError:
        logging.error(
            "Dubbed chunk volume could not be normalized. The orginial volume"
            " is used."
        )
        pass
    output_audio = output_audio.overlay(
        audio_chunk, position=start_time, loop=False
    )
  dubbed_vocals_audio_file = os.path.join(
      output_directory, AUDIO_PROCESSING, _DEFAULT_DUBBED_VOCALS_AUDIO_FILE
  )
  output_audio = output_audio.normalize()
  output_audio.export(dubbed_vocals_audio_file, format="mp3")
  return dubbed_vocals_audio_file


def merge_background_and_vocals(
    *,
    background_audio_file: str,
    dubbed_vocals_audio_file: str,
    output_directory: str,
    target_language: str,
    vocals_volume_adjustment: float = 5.0,
    background_volume_adjustment: float = 0.0,
) -> str:
  """Mixes background music and vocals tracks, normalizes the volume, and exports the result.

  Args:
      background_audio_file: Path to the background music MP3 file.
      dubbed_vocals_audio_file: Path to the vocals MP3 file.
      output_directory: Path to save the mixed MP3 file.
      target_language: The language to dub the ad into. It must be ISO 3166-1
        alpha-2 country code.
      vocals_volume_adjustment: By how much the vocals audio volume should be
        adjusted.
      background_volume_adjustment: By how much the background audio volume
        should be adjusted.

  Returns:
    The path to the output audio file with merged dubbed vocals and original
    background audio.
  """

  background = AudioSegment.from_mp3(background_audio_file)
  vocals = AudioSegment.from_mp3(dubbed_vocals_audio_file)
  background = background + background_volume_adjustment
  vocals = vocals + vocals_volume_adjustment
  shortest_length = min(len(background), len(vocals))
  background = background[:shortest_length]
  vocals = vocals[:shortest_length]
  mixed_audio = background.overlay(vocals)
  target_language_suffix = "_" + target_language.replace("-", "_").lower()
  dubbed_audio_file = os.path.join(
      output_directory,
      _OUTPUT,
      _DEFAULT_DUBBED_AUDIO_FILE
      + target_language_suffix
      + _DEFAULT_OUTPUT_FORMAT,
  )
  mixed_audio.normalize()
  mixed_audio.export(dubbed_audio_file, format="mp3")
  return dubbed_audio_file
